/*
 * Copyright  1997 The Hungry Programmers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include "wp_cont_slab.h"
#include "wp_utils.h"
#include "wp_prefs.h"

/*#define margin_debugging 1*/

#define DEBUG 1


/* a slab_extent is a struct that contains an x and y
   position as well as (left,up,right,down) values.
   They are used while doing screen updates. */

/* determine if 2 slab extents overlap */
int slab_extents_overlap (slab_extent a, slab_extent b)
{
  if (a.x - a.left >= b.x + b.right ||
      a.y - a.up >= b.y + b.down ||
      a.x + a.right <= b.x - b.left ||
      a.y + a.down <= b.y - b.up)
    return 0;

  return 1;
}

/* exchange the contents of a and b */
void slab_extents_swap (slab_extent *a, slab_extent *b)
{
  slab_extent t = (*a);
  (*a) = (*b);
  (*b) = t;
}


/* find the union of two extents, return is in orig_extent */
/* new_extent is not changed */
void slab_extent_trim (slab_extent *orig_extent, slab_extent *new_extent)
{
  if (new_extent->left < orig_extent->left)
    orig_extent->left = new_extent->left;
  if (new_extent->up < orig_extent->up)
    orig_extent->up = new_extent->up;
  if (new_extent->right < orig_extent->right)
    orig_extent->right = new_extent->right;
  if (new_extent->down < orig_extent->down)
    orig_extent->down = new_extent->down;
}


/********/
/********/
/********/


/* constructor for ContSlab */
_ContSlab::_ContSlab (WPStyle *set_style,
		      draw_ps_callback set_psdrawfunc,
		      draw_x_callback set_xdrawfunc,
		      draw_ascii_callback set_asciidrawfunc)
  : _Slab (set_style)
{
  children = NULL;

  psdrawfunc = set_psdrawfunc;
  xdrawfunc = set_xdrawfunc;
  asciidrawfunc = set_asciidrawfunc;
}


/* create a new container slab.  every slab has this virtual
   method so that code can duplicate any slab without caring
   what type it is. */
Slab _ContSlab::newOfThisType ()
{
  ContSlab c = new _ContSlab (style, psdrawfunc, xdrawfunc, asciidrawfunc);
  c->setPluginData (getPluginData ());

  c->setLeft (left);
  c->setUp (up);
  c->setRight (right);
  c->setDown (down);

  c->calculateSize (-1);

  return c;
}


/* return the direction of this's style */
int _ContSlab::getDirection ()
{
  return style->getDirection ();
}


/* insert child 'what' at position i.
   this updates the spots if needed (if a spot is in "this"
   and is at a point greater than i, it will be incremented).
   if debug >= 0, debugging spew will happen.  This returns
   true if everything works out, and false if some problem
   occures.  if a problem occures, everything should be put
   back to its original state (heh).
   The insert is accomplished by setting the child's size to 0,
   inserting it in the list of children, and then calling
   childSizeChange with the child's original size.  This causes
   the redrawing and wrapping to happen. */
boolean _ContSlab::insertChild (int i, Object what, GList *spots,
				GList *views, int debug)
{
  boolean ret;

  if (debug >= 0)
    {
      printf ("%sContSlab->insertChild this=",
	      tabber (debug));
      dd ();
      printf (" i=%d what=", i);
      ((Slab)what)->dd ();
      printf ("\n");
    }

  if (i<0 || (guint) i>g_list_length (children))
    {
      printf ("ContSlab->insert -- i=%d\n", i);
      return false;
    }

  Slab slb = (Slab) what;

  /* fix spots */
  for (int spotnum=0; (guint)spotnum<g_list_length (spots); spotnum++)
    {
      Spot s = (Spot) g_list_nth (spots, spotnum)->data;
      if (s->getSlab () == this && s->getPosition () >= i)
	s->setPosition (s->getPosition () + 1);
    }

  /* force slb to a zero size */

  int slbLeft = slb->getMLeft ();
  int slbUp = slb->getMUp ();
  int slbRight = slb->getMRight ();
  int slbDown = slb->getMDown ();

  slb->setMLeft (0);
  slb->setMUp (0);
  slb->setMRight (0);
  slb->setMDown (0);

  /* insert slb */

  children = g_list_insert (children, what, i);
  slb->setParent (this);

  /* get ready for slb's size change */
  ret = childSizeChange (slb, spots,
			 slbLeft, slbUp, slbRight, slbDown,
			 views, (debug >= 0)?debug+1:-1);

  /* set slb's size back */
  if (ret == false) /* then childSizeChange failed. */
    {
      /*children->removeElementAt (i);*/
      slb->setMLeft (slbLeft);
      slb->setMUp (slbUp);
      slb->setMRight (slbRight);
      slb->setMDown (slbDown);

      children = g_list_remove (children, what);
      calculateSize ((debug >= 0)?debug+1:-1);
      printf ("ContSlab->insertChild, childSizeChange failed.\n");
      trim_slab (spots, views, (debug >= 0)?debug+1:-1);
      return false;
    }

  if (preferences->draw_fast)
    {
      /* redraw the child that was inserted */
      WpPoint p = slb->getParent ()->getPointOfChild (slb);
      slb->draw (p.x, p.y, views, (debug >= 0)?debug+1:-1);
    }

  /* offer children to prev */
  /**/ /* XXX -- this causes endless loops */
  /*
  if (i == 0)
    {
      Slab prev = getPrevNode ();

      if (prev != NULL && prev->classID () == ContSlabID)
	{
	  ContSlab cs = (ContSlab) prev;
	  cs->attemptToStealChildren (spots, views, (debug >= 0)?debug+1:-1);
	}
    }
  */

  return true;
}




/* remove child at position i from the list of children and
   return it.  This keeps the spots from being invalidated
   (if a spot is in "this", it is moved to the next slab).
   If ignoreLostSpots is true, the spots will be
   left alone.  This is done when the calling routine knows
   that the slab will imediatly be placed back into the
   document (so the spot will not really become invalid).
   if debug >= 0, spew will happen.  The delete is done
   by setting the size of the child to 0 (which causes
   wrapping and visual update), and then removing the child
   from the list of children */
Object _ContSlab::deleteChild (int i, GList *spots, GList *views,
			       int debug, boolean ignoreLostSpots)
{
  Slab child;
  boolean ret;

  if (debug >= 0)
    {
      printf ("%sdeleteChild this=", tabber (debug));
      dd();
      printf (" i=%d\n", i);
    }

  /*if (i<0 || i>=children->size ())*/
  if (i<0 || (guint)i>=g_list_length (children))
    {
      printf ("ContSlab->delete -- i=%d\n", i);
      return NULL;
    }

  /*child = (Slab) children->elementAt (i);*/
  child = (Slab) g_list_nth (children, i)->data;

  if (style->getWhenEmpty () == WHEN_EMPTY_KEEP)
    {
      /* attempt to maintain nested structures */
      int last_in_structure = 1;
      if (child->getNextNode () && child->getNextNode ()->getParent () &&
	  child->getNextNode ()->getParent ()->getStyle () == getStyle ())
	last_in_structure = 0;
      else if (child->getPrevNode () && child->getPrevNode ()->getParent () &&
	       child->getPrevNode ()->getParent ()->getStyle () == getStyle ())
	last_in_structure = 0;

      if (last_in_structure == 1)
	{
	  printf ("not deleting because it is the last at this depth -- 0\n");
	  return NULL;
	}
    }

  /* fix the spots */

  for (int spotnum=0; (guint)spotnum<g_list_length (spots); spotnum++)
    {
      Spot s = (Spot) g_list_nth (spots, spotnum)->data;
      if (s->getSlab () == this && s->getPosition () >= i)
	s->setPosition (s->getPosition () - 1);

      if (ignoreLostSpots == false &&
	  child->slabIsInside (s->getSlab ()))
	{
	  Slab p = child->getPrevNode ();
	  if (p == NULL)
	    {
	      Slab n = child->getNextNode ();
	      if (n == NULL)
		{
		  s->setPosition (i);
		  s->setSlab (this);
		}
	      else
		{
		  s->setSlab (n);
		  s->setPosition (0);
		}
	    }
	  else
	    {
	      s->setSlab (p);
	      s->setPosition (p->getNumberOfChildren ());
	    }
	}
    }

  int saveLeft = child->getMLeft ();
  int saveUp = child->getMUp ();
  int saveRight = child->getMRight ();
  int saveDown = child->getMDown ();

  if (preferences->draw_fast)
    {
      WpPoint p = getPointOfChild (i);
      child->erase (p.x, p.y, views);
    }

  /* child becomes zero size... */
  ret = childSizeChange (child, spots,
			 - saveLeft, - saveUp, - saveRight, - saveDown,
			 views, (debug >= 0)?debug+1:-1);

  if (ret == false)
    {
      printf ("ContSlab->deleteChild, childSizeChange failed?\n");
      /*return NULL;*/
    }

  /* and delete it. */
  /*children = children->removeElementAt (i);*/
  children = g_list_remove_link (children, g_list_nth (children, i));

  /* set the size back */

  child->setMLeft (saveLeft);
  child->setMUp (saveUp);
  child->setMRight (saveRight);
  child->setMDown (saveDown);
  child->setParent (NULL);

  trim_slab (spots, views, (debug >= 0)?debug+1:-1);
  return child;
}


/* this (when finished) ensures that position i is either
   at the beginning or the end of this's list of children.
   if i is somewhere in the middle, then "this" should
   be split into 2 slabs (at i).  This is used when
   something that isn't compatible with this slab is
   being inserted in the middle of this slab.  actually,
   this is only used in wordslabs right now. */
boolean _ContSlab::ensureBorder (int i, GList *spots,
				 GList *views, int debug)
{
  ContSlab newCS;
  int num_children = g_list_length (children);

  if (debug >= 0)
    {
      printf ("%sensureBorder this=", tabber (debug));
      dd ();
      printf (" i=%d\n", i);
    }

  if (i<0 || i>num_children)
    {
      fprintf (stderr, "error -- ContSlab->ensureBorder, i=%d, ", i);
      fprintf (stderr, "contslab->num_children=%d\n", num_children);
      return false;
    }

  if (parent == NULL)
    {
      fprintf (stderr, "error -- WordSlab->ensureBorder, parent=NULL");
      return false;
    }

  if (i == 0 || i == num_children)
    return true;

  newCS = (ContSlab) newOfThisType ();


  GList *bootp;
  int n;

  for (bootp = g_list_nth (children, i), n=0;
       bootp; bootp = bootp->next, n++)
    {
      children = g_list_remove_link (children, bootp);
      Slab boot = (Slab) bootp->data;

      if (newCS->insertChild (n, boot, spots, views,
			      (debug >= 0)?debug+1:-1) == false)
	{
	  children = g_list_append (children, boot);
	  boot->setParent (this);
	  calculateSize ((debug >= 0)?debug+1:-1);
	  printf ("ContSlab->bumping -- insert failed. 0\n");
	  return false;
	}
    }

  newCS->calculateSize ((debug >= 0)?debug+1:-1);
  calculateSize ((debug >= 0)?debug+1:-1);

  /* fix the spots */
  for (int spotnum=0; (guint)spotnum<g_list_length (spots); spotnum++)
    {
      Spot s = (Spot) g_list_nth (spots, spotnum)->data;
      if (s->getSlab () == this && s->getPosition () >= i)
	{
	  s->setPosition (s->getPosition () - i);
	  s->setSlab (newCS);
	}
    }


  /* do an insert without all the redrawing, etc. */
  parent->insertChildAt (newCS, parent->getIndexOfChild (this) + 1);

  if (preferences->draw_fast)
    {
      /* redraw both children, because splitting can move them around */
      WpPoint p = getParent ()->getPointOfChild (this);
      erase (p.x, p.y, views);
      draw (p.x, p.y, views, (debug >= 0)?debug+1:-1);
      p = newCS->getParent ()->getPointOfChild (newCS);
      newCS->erase (p.x, p.y, views);
      newCS->draw (p.x, p.y, views, (debug >= 0)?debug+1:-1);
    }

  /* offer children to prev */
  if (parent->getIndexOfChild (this) == 0)
    {
      Slab prev = parent->getPrevNode ();

      if (prev != NULL && (prev->classID () == ContSlabID))
	{
	  ContSlab cs = (ContSlab) prev;
	  cs->attemptToStealChildren (spots, views, (debug >= 0)?debug+1:-1);
	}
    }

  return true;
}


/* create another slab like this one and insert it into the
   document right after "this", then return the new slab */
Slab _ContSlab::createFollower (GList *spots, GList *views, int debug)
{
  boolean ret;
  int myIndex = parent->getIndexOfChild (this);
  Slab next = newOfThisType ();

  if (debug >= 0)
    {
      printf ("%sContSlab->createFollower, this=", tabber (debug));
      dd ();
      printf ("\n");
    }

  ret = parent->insertChild (myIndex+1, next, spots,
			     views, (debug >= 0)?debug+1:-1);
  if (ret == false)
    {
      printf ("ContSlab->createFollower -- can't insert new next node.\n");
      return NULL;
    }

  return next;
}


#if 0
static void print_rectangle (int xl, int yl, int xh, int yh)
{
  printf ("   %4d        %d\n", xl, xh);
  printf ("%4d .----------. %d\n", yl, yl);
  printf ("     |  redraw  |\n");
  printf ("%4d `----------' %d\n", yh, yh);
  printf ("   %4d        %d\n", xl, xh);
}
#endif /* 0 */


/* this does a copyarea and erases the source area.  If the
   src and the dst overlap, then only part of the src area
   is erased (in order to not clobber the image) */
void _ContSlab::slideArea (GList *views, int x0, int y0, int x1, int y1,
			   int dx, int dy, int debug)
{
  if (x0 == 0 && y0 == 0 && x1 == 0 && y1 == 0)
    return;

  for (GList *v = views; v != NULL; v = v->next)
    {
      WPView *g = (WPView *) v->data;

      /*wp_view_set_color (g, wp_view_get_bg (g));*/
      wp_view_set_color (g, style->getBackground ());
      wp_view_copy_area (g, x0, y0, x1 - x0, y1 - y0, dx, dy);

      /* XXX what if the rectangles don't overlap? */

      if (debug >= 0)
	printf ("%sslide:<(%d,%d) (%d,%d)>:(%d,%d)\n",
		tabber (debug), x0, y0, x1, y1, dx, dy);

      if (dy < 0)
	wp_view_fill_rect (g, x0, y1+dy, x1-x0, -dy);
      if (dy > 0)
	wp_view_fill_rect (g, x0, y0, x1-x0, dy);
      if (dx < 0)
	wp_view_fill_rect (g, x1+dx, y0, -dx, y1-y0);
      if (dx > 0)
	wp_view_fill_rect (g, x0, y0, dx, y1-y0);

      /* XXX if the slide took material from off the edge of
	 the window, redraw some */
#     if 0 /* XXX Fix this */
      {
	int vx = wp_view_get_x (g); /* in pixels */
	int vy = wp_view_get_y (g);
	int vw = wp_view_get_width (g);
	int vh = wp_view_get_height (g);

	int xl, yl, xh, yh;

	if (dy < 0 && y1 > vy+vh)
	  {
	    xl = x0;
	    yl = vy+vh+_dy;
	    xh = x1;
	    yh = vy+vh;
	    /* XXX */
	    /*print_rectangle (xl, yl, xh, yh);*/
	  }
	if (dy > 0 && y0 < vy)
	  {
	    xl = x0;
	    yl = vy;
	    xh = x1;
	    yh = vy+dy;
	    /* XXX */
	    /*print_rectangle (xl, yl, xh, yh);*/ 
	  }
	if (dx < 0 && x1 > vx+vw)
	  {
	    xl = vx+vw+dx;
	    yl = y0;
	    xh = vx+vw;
	    yh = y1;
	    /* XXX */
	    /*print_rectangle (xl, yl, xh, yh);*/
	  }
	if (dx > 0 && x0 < vx)
	  {
	    xl = vx;
	    yl = y0;
	    xh = vx+_dx;
	    yh = y1;
	  }
      }
#     endif /* 0 */
    }
}


/* used to move a sets of containers around */
/* this tries to do slides in the right order so that */
/* none of the areas get mangled during the slides */
void _ContSlab::slide3Area (GList *views,
			    slab_extent a_orig, slab_extent a_new,
			    slab_extent b_orig, slab_extent b_new,
			    slab_extent c_orig, slab_extent c_new,
			    int debug)
{
  if (debug >= 0)
    {
      printf ("%sContSlab->slide3Area, this=", tabber (debug));
      dd ();
      printf ("\n%s  (%d %d):(%d %d %d %d) --> (%d %d):(%d %d %d %d)\n"
	      "%s  (%d %d):(%d %d %d %d) --> (%d %d):(%d %d %d %d)\n"
	      "%s  (%d %d):(%d %d %d %d) --> (%d %d):(%d %d %d %d)\n",
	      tabber (debug), a_orig.x, a_orig.y,
	      a_orig.left, a_orig.up, a_orig.right, a_orig.down,
	      a_new.x, a_new.y,
	      a_new.left, a_new.up, a_new.right, a_new.down,
	      tabber (debug), b_orig.x, b_orig.y,
	      b_orig.left, b_orig.up, b_orig.right, b_orig.down,
	      b_new.x, b_new.y,
	      b_new.left, b_new.up, b_new.right, b_new.down,
	      tabber (debug), c_orig.x, c_orig.y,
	      c_orig.left, c_orig.up, c_orig.right, c_orig.down,
	      c_new.x, c_new.y,
	      c_new.left, c_new.up, c_new.right, c_new.down);
    }

  /* attempt to arrange them so the slide happen in the right order */
  if (slab_extents_overlap (a_new, b_orig))
    {
      slab_extents_swap (&a_orig, &b_orig);
      slab_extents_swap (&a_new, &b_new);
    }
  if (slab_extents_overlap (a_new, c_orig))
    {
      slab_extents_swap (&a_orig, &c_orig);
      slab_extents_swap (&a_new, &c_new);
    }
  if (slab_extents_overlap (b_new, c_orig))
    {
      slab_extents_swap (&b_orig, &c_orig);
      slab_extents_swap (&b_new, &c_new);
    }

  /* now trim down the origs so that not too much gets copied */
  /* this is done in case a slab shrunk doing this operation */
  slab_extent_trim (&a_orig, &a_new);
  slab_extent_trim (&b_orig, &b_new);
  slab_extent_trim (&c_orig, &c_new);

  /* and do the slides */
  slideArea (views,
	     a_orig.x - a_orig.left, a_orig.y - a_orig.up,
	     a_orig.x + a_orig.right, a_orig.y + a_orig.down,
	     a_new.x - a_orig.x, a_new.y - a_orig.y, (debug >= 0)?debug+1:-1);
  slideArea (views,
	     b_orig.x - b_orig.left, b_orig.y - b_orig.up,
	     b_orig.x + b_orig.right, b_orig.y + b_orig.down,
	     b_new.x - b_orig.x, b_new.y - b_orig.y, (debug >= 0)?debug+1:-1);
  slideArea (views,
	     c_orig.x - c_orig.left, c_orig.y - c_orig.up,
	     c_orig.x + c_orig.right, c_orig.y + c_orig.down,
	     c_new.x - c_orig.x, c_new.y - c_orig.y, (debug >= 0)?debug+1:-1);
}


/* uhhh...
   this tries to pull children from the next slab into this one.
   it is called by code in wp_document. */
boolean _ContSlab::attemptToStealChildren (GList *spots,
					   GList *views,
					   int debug)
{
  Slab a_child = (Slab) g_list_nth (children, 0)->data;

  if (debug >= 0)
    {
      printf ("%sContSlab::attemptToStealChildren this=",
	      tabber (debug));
      dd ();
      printf ("\n");
    }

  return childSizeChange (a_child, spots, 0, 0, 0, 0,
			  views, (debug >= 0)?debug+1:-1);
}


/**/
/* do the visual update as a child changes its size */
/**/



/* this counts the number of children that will fit in "this".  It
   goes from the first child forward adding up the collective size
   of the children.  If it runs out of children in "this", it will
   start adding on the sizes of the children from the next node.
   It keeps going until it finds the limit.  If "this" has a
   size policy of sizebychildren, it returns the current number
   of children.  this is called csc_... becuase it is used
   by the child_size_change routine.  this also sets a flag
   that indicates weather or not the slab "child" was among
   those that will fit. */
int _ContSlab::csc_will_fit (Slab child,
			     int dLeft, int dUp, int dRight, int dDown,
			     int *child_stays,
			     int debug)
{
  int fit = 0;
  int child_girth = 0; /* size of the children, stacked end to end */
  int this_girth =
    getMajorLow (style->getDirection ()) +
    getMajorHigh (style->getDirection ());


  if (debug >= 0)
    {
      printf ("%sContSlab::csc_will_fit this=",
	      tabber (debug));
      dd ();
      printf (" child=");
      child->dd ();
    }

  (*child_stays) = 0;

  Slab p = NULL; /* the slab before s */

  for (Slab s = (Slab) g_list_nth (children, 0)->data;
       s != NULL;
       s = s->getNextNode (), fit++, p = s)
    {
      child_girth += (s->getMMajorLow (style->getDirection ()) +
		      s->getMMajorHigh (style->getDirection ()));
	  
      if (s == child)
	{
	  switch (style->getDirection ())
	    {
	    case DIR_LEFT:
	    case DIR_RIGHT:
	      child_girth += (dLeft + dRight);
	      break;
	    case DIR_UP:
	    case DIR_DOWN:
	      child_girth += (dUp + dDown);
	      break;
	    }
	}

      /* if s is in a different type of parent, don't pull
	 it into this one */
      if (s->getParent () && s->getParent ()->getStyle () != getStyle ())
	break;

      /* if s needs to be first, and isn't, it wont fit here */
      if (s != (Slab) g_list_nth (children, 0)->data && s->isFirst ())
	break;

      /* if the previous slab was last, s wont fit */
      if (p != NULL && p->isLast ())
	break;

      /* if the children's cumulative size is too great, s wont fit */
      if (getMajorSizePolicy () != SIZE_POLICY_BY_CHILDREN &&
	  (child_girth > this_girth))
	/* unless s is a space */
	if (s->classID () != STSlabID)
	  break;

      if (s == child)
	(*child_stays) = 1;
    }

  if (debug >= 0)
    printf (" fit:%d child_stays:%s\n", fit,
	    (*child_stays)?"true":"false");

  return fit;
}


/* returns the largest of a and b */
long int csc_max (long int a, long int b)
{
  if (a > b)
    return a;
  return b;
}

/* returns the smallest of a and b */
long int csc_min (long int a, long int b)
{
  if (a < b)
    return a;
  return b;
}

/* find the smallest rectangle that will cover the children
   between c0 and c1.  inc_c0 and inc_c1 determine whether
   the ends of this region are inclusive (!0) or exclusive (0).
   The coords of the returned rectangle are relative to the
   position of "this" */
slab_extent
_ContSlab::csc_cover_children (Slab c0, Slab c1, int inc_c0, int inc_c1)
{
  slab_extent r;
  WpPoint myPoint;

  if (parent != NULL)
    myPoint = parent->getPointOfChild (this);
  else
    myPoint = newWpPoint (getMLeft (), getMUp ());

  r.left = r.up = r.right = r.down = 0;

  if ((!c0) || (!c1)) return r;

  if ((c0 == c1 && ((! inc_c0) || (! inc_c1))) ||
      (c0->getNextNode () == c1) && ((! inc_c0) && (! inc_c1)))
    {
      r.x = r.y = 0;
      return r;
    }

  if (! inc_c0) c0 = c0->getNextNode ();
  if (! inc_c1) c1 = c1->getPrevNode ();

  if ((!c0) || (!c1)) return r;

  /* something to start with */
  WpPoint t = getPointOfChild (c0);
  r.x = t.x;
  r.y = t.y;

  for (Slab s=c0; s != c1->getNextNode (); s=s->getNextNode ())
    {
      if (s == NULL) return r;

      WpPoint s_point = getPointOfChild (s);

      r.left = csc_max (r.left, r.x - (s_point.x - s->getMLeft ()));
      r.up = csc_max (r.up, r.y - (s_point.y - s->getMUp ()));
      r.right = csc_max (r.right, (s_point.x + s->getMRight ()) - r.x);
      r.down = csc_max (r.down, (s_point.y + s->getMDown ()) - r.y);
    }

  /* make the coords relative */
  r.x -= myPoint.x;
  r.y -= myPoint.y;

  /* XXX include the margins */
  switch (style->getDirection ())
    {
    case DIR_LEFT:
    case DIR_RIGHT:
      r.up = getMUp ();
      r.down = getMDown ();
      break;
    case DIR_UP:
    case DIR_DOWN:
      r.left = getMLeft ();
      r.right = getMRight ();
      break;
    }

  /* XXX include the end margins */

  if (c0 == (Slab) (children->data)) /* c0 is the first child */
    switch (style->getDirection ())
      {
      case DIR_LEFT:
	r.right += style->getMarginRight () + 1;
	break;
      case DIR_RIGHT:
	r.left += style->getMarginLeft () + 1;
	break;
      case DIR_UP:
	r.down += style->getMarginDown () + 1;
	break;
      case DIR_DOWN:
	r.up += style->getMarginUp () + 1;
	break;
      }

#if 1
  if (c1 == (Slab) (g_list_last (children)->data)) /* c1 is the last child */
    switch (style->getDirection ())
     {
      case DIR_LEFT:
	r.left += style->getMarginLeft () + 1;
	break;
      case DIR_RIGHT:
	r.right += style->getMarginRight () + 1;
	break;
      case DIR_UP:
	r.up += style->getMarginUp () + 1;
	break;
      case DIR_DOWN:
	r.down += style->getMarginDown () + 1;
	break;
      }
#endif /* 0 */

  /* ok */


  return r;
}


/* this pushes children into the next slab until the number
   of children in "this" is equal to "numFit". */
boolean
_ContSlab::csc_boot_children (GList *spots, int numFit,
			      GList *views, int debug)
{
  if (debug >= 0)
    {
      printf ("%scsc_boot_children this=",
	      tabber (debug));
      dd ();
      printf (" numFit=%d", numFit);
      printf ("\n");
    }

  /*** push out children that wont fit ***/

  while ((guint)numFit < g_list_length (children) && numFit > 0)
    {
      int n = g_list_length (children) - 1;
      Slab boot = (Slab) getChildAt (n);
      Slab next = getNextNode ();
      WpPoint p = getPointOfChild (n);
      if (preferences->draw_fast)
	boot->erase (p.x, p.y, views);

      children = g_list_remove_link (children, g_list_nth (children, n));


      if (next == NULL || next->classID () != ContSlabID)
	next = createFollower (spots, views, (debug >= 0)?debug+1:-1);

      if (next == NULL)
	{
	  children = g_list_append (children, boot);
	  boot->setParent (this);
	  printf ("ContSlab->bumping -0- no follower.\n");
	  return false;
	}

      /* avoid the scenario where every lines */
      /*   gives it's children to the line after it */
      if (next->classID () == ContSlabID &&
	  next->getNumberOfChildren () > 0 &&
	  ((Slab) next->getChildAt (0))->isFirst () > 0)
	next = createFollower (spots, views, (debug >= 0)?debug+1:-1);

      if (next == NULL)
	{
	  children = g_list_append (children, boot);
	  boot->setParent (this);
	  printf ("ContSlab->bumping -1- no follower.\n");
	  return false;
	}

      if (debug >= 0)
	{
	  printf ("%sPushing Child, this=", tabber (debug));
	  dd ();
	  printf (" what=");
	  boot->dd ();
	  printf ("\n");
	}

      if (next->insertChild (0, boot, spots, views,
			     (debug >= 0)?debug+1:-1) == false)
	{
	  children = g_list_append (children, boot);
	  boot->setParent (this);
	  calculateSize ((debug >= 0)?debug+1:-1);
	  printf ("ContSlab->bumping -- insert failed. 1\n");
	  return false;
	}

      /* fix the spots */
      for (int spotnum=0; (guint)spotnum<g_list_length (spots); spotnum++)
	{
	  Spot s = (Spot) g_list_nth (spots, spotnum)->data;
	  if (s->getSlab () == this &&
	      s->getPosition () >= (int) g_list_length (children))
	    {
	      s->setPosition (0);
	      s->setSlab (next);
	    }
	}
    }

  return true;
}


/* this steals children from the next slab until the number
   of children in "this" is equal to "numFit". */
boolean
_ContSlab::csc_steal_children (GList *spots, int numFit,
			       GList *views, int debug)
{
  /*** pull in new  children ***/

  while ((guint)numFit > g_list_length (children))
    {
      /*int n = g_list_length (children);*/
      Slab steal;
      Slab next = getNextNode ();

      if (next == NULL)
	return false;

      if ((steal = (Slab) next->deleteChild (0, spots, views,
					     (debug >= 0)?debug+1:-1,
					     true)) == NULL)
	/* XXX or false?  ignore lost spots? */
	{
	  printf ("ContSlab->stealing -- delete failed.\n");
	  return false;
	}

      if (debug >= 0)
	{
	  printf ("%sPulling Child, this=", tabber (debug));
	  dd ();
	  printf (" what=");
	  steal->dd ();
	  printf ("\n");
	}

      children = g_list_append (children, steal);
      steal->setParent (this);

      /*
      WpPoint p = getPointOfChild (n);
      boot->draw (p.x, p.y, views, (debug >= 0)?debug+1:-1);
      */
    }

  return true;
}


/* convert a rectangle with relative coords to one with absolute coords */
slab_extent _ContSlab::csc_rect_relative_to_absolute (slab_extent in)
{
  /* get the current position of "this" */
  WpPoint myPoint;
  if (parent != NULL)
    myPoint = parent->getPointOfChild (this);
  else
    myPoint = newWpPoint (getMLeft (), getMUp ());

  in.x += myPoint.x;
  in.y += myPoint.y;

  return in;
}


/* this remembers the starting size of "this", calls calculateSize,
   and checks to see if they are different.  if they are, it
   sets the size back to the original, and calls
   parent->childSizeChange.
   XXX this code looks a lot like _Slab::styleChangeNotify,
   they could probably be combined */
boolean _ContSlab::csc_this_size_might_change (GList *spots, GList *views,
					       int debug)
{
  int save_left = left;
  int save_up = up;
  int save_right = right;
  int save_down = down;
  int new_left, new_up, new_right, new_down;


  if (debug >= 0)
    {
      printf ("%scsc_this_size_might_change this=",
	      tabber (debug));
      dd ();
      printf ("\n");
    }

  calculateSize ((debug >= 0)?debug+1:-1);

  if (left == save_left &&
      up == save_up &&
      right == save_right &&
      down == save_down)
    return true;

  if (parent == NULL)
    return true;

  new_left = left;
  new_up = up;
  new_right = right;
  new_down = down;
  left = save_left;
  up = save_up;
  right = save_right;
  down = save_down;

  return parent->childSizeChange (this, spots,
				  new_left - left,
				  new_up - up,
				  new_right - right,
				  new_down - down,
				  views, (debug >= 0)?debug+1:-1);
}


/* expanding child stays, some leave */
boolean _ContSlab::csc_boot_some (Slab child, GList *spots,
				  int dLeft, int dUp, int dRight, int dDown,
				  int fit,
				  GList *views, int debug)
{
  if (debug >= 0)
    {
      printf ("%scsc_boot_some this=",
	      tabber (debug));
      dd ();
      printf (" dSize=(%d,%d,%d,%d)", dLeft, dUp, dRight, dDown);
      printf (" fit=%d", fit);
      printf (" child=");
      child->dd ();
      printf ("\n");
    }

  WpPoint myc;
  Slab first, last;
  slab_extent rl0, rc0, rh0;

  if (preferences->draw_fast)
    {
      first = (Slab) g_list_nth (children, 0)->data;
      last = (Slab) g_list_nth (children, fit-1)->data;

      /* chop the children up into 3 rectangles */
      rl0 = csc_cover_children (first, child, 1, 0);
      rc0 = csc_cover_children (child, child, 1, 1);
      rh0 = csc_cover_children (child, last, 0, 1);
    }

  if (parent != NULL)
    myc = parent->getPointOfChild (this);
  else
    {
      myc.x = getMLeft ();
      myc.y = getMUp ();
    }
  WpPoint cc0 = getPointOfChild (child);
  cc0.x -= myc.x;
  cc0.y -= myc.y;

  if (csc_boot_children (spots, fit, views, debug) == false)
    return false;

  child->setMLeft (child->getMLeft () + dLeft);
  child->setMUp (child->getMUp () + dUp);
  child->setMRight (child->getMRight () + dRight);
  child->setMDown (child->getMDown () + dDown);

  if (preferences->draw_fast)
    {
      slab_extent rl1 = csc_cover_children (first, child, 1, 0);
      slab_extent rc1 = csc_cover_children (child, child, 1, 1);
      slab_extent rh1 = csc_cover_children (child, last, 0, 1);

	/*
	if (parent != NULL)
	  myc = parent->getPointOfChild (this);
	else
	  {
	    myc.x = getMLeft ();
	    myc.y = getMUp ();
	  }
	WpPoint cc1 = getPointOfChild (child);
	cc1.x -= myc.x;
	cc1.y -= myc.y;
	*/  

      rl0 = csc_rect_relative_to_absolute (rl0);
      rc0 = csc_rect_relative_to_absolute (rc0);
      rh0 = csc_rect_relative_to_absolute (rh0);
      rl1 = csc_rect_relative_to_absolute (rl1);
      rc1 = csc_rect_relative_to_absolute (rc1);
      rh1 = csc_rect_relative_to_absolute (rh1);
      
      slide3Area (views, rl0, rl1, rc0, rc1, rh0, rh1,
		  (debug >= 0)?debug+1:-1);
    }

  if (csc_this_size_might_change (spots, views,
				  (debug >= 0)?debug+1:-1) == false)
    return false;

  return true;
}


/* expanding child stays, get some children from next*/
boolean _ContSlab::csc_steal_some (Slab child, GList *spots,
				   int dLeft, int dUp, int dRight, int dDown,
				   int fit,
				   GList *views, int debug)
{
  if (debug >= 0)
    {
      printf ("%scsc_steal_some this=",
	      tabber (debug));
      dd ();
      printf (" dSize=(%d,%d,%d,%d)", dLeft, dUp, dRight, dDown);
      printf (" fit=%d", fit);
      printf (" child=");
      child->dd ();
      printf ("\n");
    }

  WpPoint myc;
  int orig_num_children;
  Slab first, last;
  slab_extent rl0, rc0, rh0;

  if (preferences->draw_fast)
    {
      orig_num_children = g_list_length (children);

      first = (Slab) g_list_nth (children, 0)->data;
      last = (Slab) g_list_nth (children, orig_num_children-1)->data;

      /* chop the children up into 3 rectangles */
      rl0 = csc_cover_children (first, child, 1, 0);
      rc0 = csc_cover_children (child, child, 1, 1);
      rh0 = csc_cover_children (child, last, 0, 1);
    }

  if (parent != NULL)
    myc = parent->getPointOfChild (this);
  else
    {
      myc.x = getMLeft ();
      myc.y = getMUp ();
    }
  WpPoint cc0 = getPointOfChild (child);
  cc0.x -= myc.x;
  cc0.y -= myc.y;

  if (csc_steal_children (spots, fit, views, debug) == false)
    return false;

  child->setMLeft (child->getMLeft () + dLeft);
  child->setMUp (child->getMUp () + dUp);
  child->setMRight (child->getMRight () + dRight);
  child->setMDown (child->getMDown () + dDown);

  if (csc_this_size_might_change (spots, views,
				  (debug >= 0)?debug+1:-1) == false)
    return false;

  if (preferences->draw_fast)
    {
      slab_extent rl1 = csc_cover_children (first, child, 1, 0);
      slab_extent rc1 = csc_cover_children (child, child, 1, 1);
      slab_extent rh1 = csc_cover_children (child, last, 0, 1);

      if (parent != NULL)
	myc = parent->getPointOfChild (this);
      else
	{
	  myc.x = getMLeft ();
	  myc.y = getMUp ();
	}
      WpPoint cc1 = getPointOfChild (child);
      cc1.x -= myc.x;
      cc1.y -= myc.y;

      rl0 = csc_rect_relative_to_absolute (rl0);
      rc0 = csc_rect_relative_to_absolute (rc0);
      rh0 = csc_rect_relative_to_absolute (rh0);
      rl1 = csc_rect_relative_to_absolute (rl1);
      rc1 = csc_rect_relative_to_absolute (rc1);
      rh1 = csc_rect_relative_to_absolute (rh1);

      slide3Area (views, rl0, rl1, rc0, rc1, rh0, rh1,
		  (debug >= 0)?debug+1:-1);

      /* draw the new children */
      for (GList *i = g_list_nth (children, orig_num_children-1)->next;
	   i != NULL;
	   i = i->next)
	{
	  Slab ichild = (Slab) i->data;
	  WpPoint p = getPointOfChild (ichild);
	  ichild->draw (p.x, p.y, views, (debug >= 0)?debug+1:-1);
	}
    }

  return true;
}


/* expanding child leaves */
boolean _ContSlab::csc_boot_child (Slab child, GList *spots,
				   int dLeft, int dUp, int dRight, int dDown,
				   int fit,
				   GList *views, int debug)

{
  if (debug >= 0)
    {
      printf ("%scsc_boot_child this=",
	      tabber (debug));
      dd ();
      printf (" dSize=(%d,%d,%d,%d)", dLeft, dUp, dRight, dDown);
      printf (" fit=%d", fit);
      printf (" child=");
      child->dd ();
      printf ("\n");
    }

  Slab first;
  Slab last;

  /* remember the rectangle which encompasses the slabs before "child" */
  slab_extent rl0;
  slab_extent rch0;

  if (preferences->draw_fast)
    {
      first = (Slab) g_list_nth (children, 0)->data;
      last = (Slab) g_list_nth (children, fit-1)->data;

      /* remember the rectangle which encompasses the slabs before "child" */
      rl0 = csc_cover_children (first, child, 1, 0);
      rch0 = csc_cover_children (child, last, 1, 1);

      rch0 = csc_rect_relative_to_absolute (rch0);

      /* erase the slabs that are getting bumped */
      for (GList *v = views; v != NULL; v = v->next)
	wp_view_fill_rect ((WPView *) v->data,
			   rch0.x-rch0.left, rch0.y-rch0.up,
			   rch0.left + rch0.right,
			   rch0.up + rch0.down);
    }

  child->setMLeft (child->getMLeft () + dLeft);
  child->setMUp (child->getMUp () + dUp);
  child->setMRight (child->getMRight () + dRight);
  child->setMDown (child->getMDown () + dDown);

  if (csc_boot_children (spots, fit, views, (debug >= 0)?debug+1:-1) == false)
    return false;

  if (preferences->draw_fast)
    {
      slab_extent rl1 = csc_cover_children (first, child, 1, 0);

      rl0 = csc_rect_relative_to_absolute (rl0);
      rl1 = csc_rect_relative_to_absolute (rl1);

      slideArea (views,
		 rl0.x - rl0.left, rl0.y - rl0.up,
		 rl0.x + rl0.right, rl0.y + rl0.down,
		 rl1.x - rl0.x, rl1.y - rl0.y, (debug >= 0)?debug+1:-1);
    }

  if (csc_this_size_might_change (spots, views,
				  (debug >= 0)?debug+1:-1) == false)
    return false;

  return true;
}


/* this is the entry point for wrapping changes.  it calls
   csc_will_fit in order to determine how many children
   *should* be in "this".  It then calls routines that will
   push some out or steals some from the next slab so that
   the number of children is correct.  This is called
   when a child is going to change its size (but after
   the change in the child that will causes it's size to
   change has been completed -- example: insert the character,
   into the wordslab, but don't change its size, *then* call this. */
boolean _ContSlab::childSizeChange (Slab child, GList *spots,
				    /* change in child */
				    int dLeft, int dUp, int dRight, int dDown,
				    GList *views, int debug)
{
  if (debug >= 0)
    {
      printf ("%schildSizeChange this=",
	      tabber (debug));
      dd ();
      printf (" dSize=(%d,%d,%d,%d)", dLeft, dUp, dRight, dDown);
      printf (" child=");
      child->dd ();
      printf ("\n");
    }

  if (g_list_length (children) == 0)
    {
      fprintf (stderr, "childSizeChange -- no children?\n");
      return false;
    }

  int child_stays;
  int fit = csc_will_fit (child, dLeft, dUp, dRight, dDown,
			  &child_stays,
			  (debug >= 0)?debug+1:-1);

  if (fit == 0)
    {
      /*fprintf (stderr, "childSizeChage -- booting all children?\n");*/
      /* if we end up here, the user has probably typed a word that
	 is longer than a line.  I'm not sure how to handle this... */
      return false;
    }

  int ret;

  if (child_stays)
    {
      int current_number = g_list_length (children);

      if (current_number >= fit)
	ret = csc_boot_some (child, spots, dLeft, dUp, dRight, dDown,
			     fit, views, (debug >= 0)?debug+1:-1);
      else
	ret = csc_steal_some (child, spots, dLeft, dUp, dRight, dDown,
			      fit, views, (debug >= 0)?debug+1:-1);
    }
  else
    ret = csc_boot_child (child, spots, dLeft, dUp, dRight, dDown,
			  fit, views, (debug >= 0)?debug+1:-1);

  return ret;
}


/* returns the first leaf slab contained in this or any child of this */
Slab _ContSlab::firstLowestContainer ()
{
  if (g_list_length (children) == 0)
    return this;

  Slab child = (Slab) g_list_first (children)->data;
  return child->firstLowestContainer ();
}


/* returns the last leaf slab contained in this or any child of this */
Slab _ContSlab::lastLowestContainer ()
{
  if (g_list_length (children) == 0)
    return this;

  Slab child = (Slab) g_list_last (children)->data;
  return child->lastLowestContainer ();
}

/* a parent can ask a child if it prefers to be the first
   slab in the parents list.  the return value is the
   number of levels up that this is true.  if a slab
   returns 0, it doesn't care if it is first.  if it
   return 2, it wants to be the first child in its parent's list,
   and it wants its parent to be the first child in its
   parent's list */
int _ContSlab::isFirst ()
{
  int r;

  if (g_list_length (children) == 0) return 0;

  Slab firstChild = (Slab) g_list_first (children)->data;
  r = firstChild->isFirst () - 1;

  if (r > 0) return r - 1;

  return 0;
}

/* a parent can ask a child if it prefers to be the last
   slab in the parents list.  the return value is the
   number of levels up that this is true.  if a slab
   returns 0, it doesn't care if it is last.  if it
   return 2, it wants to be the last child in its parent's list,
   and it wants its parent to be the last child in its
   parent's list */
int _ContSlab::isLast ()
{
  int r;

  if (g_list_length (children) == 0) return 0;

  Slab lastChild = (Slab) g_list_last (children)->data;
  r = lastChild->isLast () - 1;

  if (r > 0) return r - 1;

  return 0;
}


/* this is used to move "sideways" in the tree structure.
   It will return the child in this's parent's list of children
   that follows this.  If this is the last child in this's
   parent's list, then it returns the first child contained
   in the slab that follow this's parent.  it will return
   NULL if the next slab can't be found. */
Slab _ContSlab::getNextNode ()
{
  if (parent == NULL)
    return NULL;

  return parent->getNextNodeUp (this, 1);
}

/* this is used by getNextNode -- don't call it directly. */
Slab _ContSlab::getNextNodeUp (Slab child, int depth)
{
  int childIndex = getIndexOfChild (child);

  if (childIndex < 0)
    return NULL;

  if ((guint)childIndex == g_list_length (children) - 1)
    {
      if (parent == NULL)
	return NULL;

      return parent->getNextNodeUp (this, depth + 1);
    }

  Slab nn = (Slab) g_list_nth (children, childIndex + 1)->data;

  if (depth == 0)
    return nn;
  else
    return nn->getNextNodeDown (depth - 1);
}

/* this is used by getNextNode -- don't call it directly. */
Slab _ContSlab::getNextNodeDown (int depth)
{
  if (depth == 0)
    return this;

  if (g_list_length (children) == 0)
    {
      Slab nn = getNextNode ();
      if (nn == NULL)
	{
	  printf ("ContSlab->nextNodeDown -- no children.\n");
	  return NULL;
	}

      return nn->getNextNodeDown (depth);
    }

  Slab fn = (Slab) g_list_first (children)->data;

  return fn->getNextNodeDown (depth - 1);
}


/* this is used to move "sideways" in the tree structure.
   It will return the child in this's parent's list of children
   that precedes this.  If this is the first child in this's
   parent's list, then it returns the last child contained
   in the slab that precedes this's parent.  it will return
   NULL if the previous slab can't be found. */
Slab _ContSlab::getPrevNode ()
{
  if (parent == NULL)
    return NULL;

  return parent->getPrevNodeUp (this, 1);
}


/* this is used by getPrevNode -- don't call it directly. */
Slab _ContSlab::getPrevNodeUp (Slab child, int depth)
{
  int childIndex = getIndexOfChild (child);

  if (childIndex < 0)
    return NULL;

  if (childIndex == 0)
    {
      if (parent == NULL)
	return NULL;

      return parent->getPrevNodeUp (this, depth + 1);
    }

  Slab pn = (Slab) g_list_nth (children, childIndex - 1)->data;

  if (depth == 0)
    return pn;
  else
    return pn->getPrevNodeDown (depth - 1);
}


/* this is used by getPrevNode -- don't call it directly. */
Slab _ContSlab::getPrevNodeDown (int depth)
{
  if (depth == 0)
    return this;

  if (g_list_length (children) == 0)
    {
      Slab pn = getNextNode ();
      if (pn == NULL)
	{
	  printf ("ContSlab::prevNodeDown -- no children.\n");
	  return NULL;
	}

      return pn->getPrevNodeDown (depth);
    }

  Slab ln = (Slab) g_list_last (children)->data;
  return ln->getPrevNodeDown (depth - 1);
}


/* returns the i'th child from the list of this's children. */
Object _ContSlab::getChildAt (int i)
{
  if (g_list_length (children) < (guint)i)
    {
      printf ("ContSlab->getChildAt -- error -- %d > %d\n",
	      i, g_list_length (children));
      return NULL;
    }

  return (Object) g_list_nth (children, i)->data;
}

/* this inserts a "what" into the list of children at position
   i, but doesn't update the screen or cause text wrapping. */
void _ContSlab::insertChildAt (Object what, int i)
{
  /*children->insertElementAt (what, i);*/
  g_list_insert (children, what, i);
  ((Slab) what)->setParent (this);
}


/* this is used for debugging.  it does a recursive screen
   dump of the structure of the slab tree.  tab is incremented
   as the depth of the call increases in order to make the
   output more pleasing to the eye */
void _ContSlab::dumpToScreen (int tab)
{
  WPSizePolicy xsp, ysp;
  xsp = style->getXSizePolicy ();
  ysp = style->getYSizePolicy ();

  for (int t=0; t<tab; t++) printf ("+-");

  switch (getDepth ())
    {
    case 1: printf ("Doc"); break;
    case 2: printf ("Page"); break;
    case 3: printf ("Col"); break;
    case 4: printf ("Line"); break;
    default: printf ("Cont"); break;
    }

  printf ("(%d %d %d %d)", left, up, right, down);
  printf ("(%ld %ld %ld %ld) ",
	  style->getMarginLeft (),
	  style->getMarginUp (),
	  style->getMarginRight (),
	  style->getMarginDown ());
  
  printf ("dir:%s ",
	  ((style->getDirection ()==DIR_LEFT)?"L":
	   (style->getDirection ()==DIR_UP)?"U":
	   (style->getDirection ()==DIR_RIGHT)?"R":
	   (style->getDirection ()==DIR_DOWN)?"D":"bogus"));
  printf ("xsp:%s ",
	  ((xsp==SIZE_POLICY_FIXED)?"fixed":
	   (xsp==SIZE_POLICY_BY_CHILDREN)?"chldrn":
	   (xsp==SIZE_POLICY_BY_PARENT)?"prnt":
	   "bogus"));
  printf ("ysp:%s ",
	  ((ysp==SIZE_POLICY_FIXED)?"fixed":
	   (ysp==SIZE_POLICY_BY_CHILDREN)?"chldrn":
	   (ysp==SIZE_POLICY_BY_PARENT)?"prnt":
	   "bogus"));

  printf ("pp:%p", plugin_data);

  printf (" <this=%p> <parent=%p>\n", this, parent);

  for (int i=0; (guint)i<g_list_length (children); i++)
    {
      Slab child = (Slab) g_list_nth (children, i)->data;
      child->dumpToScreen (tab + 1);
    }
}


/* this does a recursize dump of the slab tree to the screen, but
   keeps the output on one line */
void _ContSlab::dd ()
{
  Slab child;

  int i;
  int d = getDepth ();

  printf ("{");

  if (d == 1)
    {
      printf ("<<<Document>>>");
      return;
    }

  if (d == 2)
    {
      printf ("<<<Page>>>");
      return;
    }

  if (d == 3)
    {
      printf ("<<<Column>>>");
      return;
    }

  for (i=0; (guint)i<g_list_length (children); i++)
    {
      child = (Slab) g_list_nth (children, i)->data;
      child->dd ();
    }

  printf ("}");
}


/* given a child, this will return its index into the list
   of this's children, or -1 if "what" is not a child of "this" */
int _ContSlab::getIndexOfChild (Object what)
{
  int i = 0;

  for (GList *gl = children; gl != NULL; gl = gl->next, i++)
    if (gl->data == what)
      return i;

  fprintf (stderr, "ContSlab -- getIndexOfChild -- child not found\n");

  return -1;

  /*return children->indexOf (what);*/
}


/* return the number of children in this's list of children */
int _ContSlab::getNumberOfChildren ()
{
  return g_list_length (children);
}


/* this sets up, left, right, and down to what they should
   be given the current children and style settings.
   if an axis has a style setting of fixed or size by
   parent, the values along that axis wont be changed.
   if an axis is size-by-children, then the children
   are stacked along that axis and their sizes are summed up. */
void _ContSlab::calculateSize (int debug)
{
  int saveLeft=left, saveUp=up, saveRight=right, saveDown=down;
  WPSizePolicy xsp, ysp;
  xsp = style->getXSizePolicy ();
  ysp = style->getYSizePolicy ();

  if (debug >= 0)
    {
      printf ("%scalculateSize this=", tabber (debug));
      dd ();
      printf ("(l,u,r,d):%d %d %d %d ", left, up, right, down);
    }

  if (xsp == SIZE_POLICY_BY_PARENT)
    {
      if (parent == NULL)
	{
	  printf ("ContSlab->calculateSize, ");
	  printf ("xSizePolicy == SIZE_POLICY_BY_PARENT, ");
	  printf ("but parent is NULL\n");
	  return;
	}
      left = parent->getMLeft ();
      right = parent->getMRight ();
    }
  if (ysp == SIZE_POLICY_BY_PARENT)
    {
      if (parent == NULL)
	{
	  printf ("ContSlab->calculateSize, ");
	  printf ("ySizePolicy == SIZE_POLICY_BY_PARENT, ");
	  printf ("but parent is NULL\n");
	  return;
	}
      up = parent->getMUp ();
      down = parent->getMDown ();
    }

  if (xsp == SIZE_POLICY_BY_CHILDREN ||
      ysp == SIZE_POLICY_BY_CHILDREN)
    {
      if (xsp == SIZE_POLICY_BY_CHILDREN)
	left = right = 0;
      if (ysp == SIZE_POLICY_BY_CHILDREN)
	up = down = 0;

      for (int i=0; (guint)i<g_list_length (children); i++)
	{
	  Slab child = (Slab) g_list_nth (children, i)->data;

	  if (xsp == SIZE_POLICY_BY_CHILDREN)
	    {
	      if (style->getDirection () == DIR_LEFT)
		left += (child->getMLeft () + child->getMRight ());
	      else if (style->getDirection () == DIR_RIGHT)
		right += (child->getMLeft () + child->getMRight ());
	      else
		{
		  if (left < child->getMLeft ()) left = child->getMLeft ();
		  if (right < child->getMRight ()) right = child->getMRight ();
		}
	    }

	  if (ysp == SIZE_POLICY_BY_CHILDREN)
	    {
	      if (style->getDirection () == DIR_UP)
		up += (child->getMUp () + child->getMDown ());
	      else if (style->getDirection () == DIR_DOWN)
		down += (child->getMUp () + child->getMDown ());
	      else
		{
		  if (up < child->getMUp ()) up = child->getMUp ();
		  if (down < child->getMDown ()) down = child->getMDown ();
		}
	    }
	}
    }

  /* if the size of this slab changed, let children resize themselves */

  if (saveLeft != left || saveUp != up ||
      saveRight != right || saveDown != down)
    {
      for (int i=0; (guint)i<g_list_length (children); i++)
	{
	  Slab child = (Slab) g_list_nth (children, i)->data;

	  if (child->getXSizePolicy () == SIZE_POLICY_BY_PARENT ||
	      child->getYSizePolicy () == SIZE_POLICY_BY_PARENT)
	    child->calculateSize ((debug >= 0)?debug+1:-1);
	}
    }

  if (style->getXSizePolicy () == SIZE_POLICY_FIXED)
    {
      left = style->getFixedLeft ();
      right = style->getFixedRight ();
    }
  if (style->getYSizePolicy () == SIZE_POLICY_FIXED)
    {
      up = style->getFixedUp ();
      down = style->getFixedDown ();
    }

  if (debug >= 0)
    printf (" ---> %d %d %d %d\n", left, up, right, down);
}


/* this recurses upward in the tree to find the x/y offset
   of "child" from the upper left corner of the document.
   this has nothing to do with scrollbar settings or anything
   the user is looking at. */
WpPoint _ContSlab::getPointOfChild (Slab child)
{
  return getPointOfChild (getIndexOfChild (child));
}


/* this recurses upward in the tree to find the x/y offset
   of "child" from the upper left corner of the document.
   this has nothing to do with scrollbar settings or anything
   the user is looking at. */
WpPoint _ContSlab::getPointOfChild (int childNum)
{
  WpPoint myPoint;
  int x, y;
  Slab child;

  if (parent != NULL)
    {
      myPoint = parent->getPointOfChild (this);
      x = myPoint.x;
      y = myPoint.y;
    }
  else
    {
      x = getMLeft ();
      y = getMUp ();
    }

# if DEBUG
  if (childNum < 0 || (guint)childNum > g_list_length (children))
    {
      printf ("ContSlab->getPointOfChild -- error -- ");
      printf ("childNum=%d\n", childNum);
      return newWpPoint (x, y);
    }
# endif /* DEBUG */

  /* skip over the children before childNum */
  int i=0;
  GList *childp;
  WPDirection sd = style->getDirection ();
  for (childp=children; i<childNum; i++, childp=childp->next)
    {
      child = (Slab) childp->data;

      switch (sd)
	{
	case DIR_LEFT: x -= child->getMWidth (); break;
	case DIR_UP: y -= child->getMHeight (); break;
	case DIR_RIGHT: x += child->getMWidth (); break;
	case DIR_DOWN: y += child->getMHeight (); break;
	}
    }


  if (childp != NULL)
    {
      /* skip to the center of the childnum child */
      /*child = (Slab) g_list_nth (children, childNum)->data;*/
      child = (Slab) childp->data;

      switch (sd)
	{
	case DIR_LEFT:
	  x -= child->getMRight ();
	  /*y -= style->getMarginDown ();*/
	  /*y += child->getUp ();*/
	  break;
	case DIR_UP:
	  /*x -= style->getMarginRight ();*/
	  /*x += child->getLeft ();*/
	  y -= child->getMDown ();
	  break;
	case DIR_RIGHT:
	  x += child->getMLeft ();
	  /*y += style->getMarginUp ();*/
	  /*y += child->getUp ();*/
	  break;
	case DIR_DOWN:
	  /*x += style->getMarginLeft ();*/
	  /*x += child->getLeft ();*/
	  y += child->getMUp ();
	  break;
	}
    }

  /* deal with centering */
  alignmentAdd (&x, &y);

  return newWpPoint (x, y);
}


/* given an offset from the upper left corner of the document,
   return a Spot in a leaf slab.  this is used mainly for
   when a user clicks on a document with the mouse */
Spot _ContSlab::getSpotOfPoint (int px, int py)
{
  int num = g_list_length (children);
  int x=0, y=0;
  int kx=0, ky=0;
  Slab child=NULL;

  if (num == 0)
    {
      printf ("ContSlab->getSpotOfPoint -- no children.\n");
      return NULL;
    }

  alignmentAdd (&x, &y);

  /* try to find child that contains the click */
  for (int i=0; i<num; i++)
    {
      child = (Slab) g_list_nth (children, i)->data;

      switch (style->getDirection ())
	{
	case DIR_LEFT:
	  x -= child->getMRight ();
	  break;
	case DIR_UP:
	  y -= child->getMDown ();
	  break;
	case DIR_RIGHT:
	  x += child->getMLeft ();
	  break;
	case DIR_DOWN:
	  y += child->getMUp ();
	  break;
	}

      kx = x;
      ky = y;

      switch (style->getDirection ())
	{
	case DIR_LEFT: x -= child->getMLeft (); break;
	case DIR_UP: y -= child->getMUp (); break;
	case DIR_RIGHT: x += child->getMRight (); break;
	case DIR_DOWN: y += child->getMDown (); break;
	}

      switch (style->getDirection ())
	{
	case DIR_LEFT:
	  if (x <= px) return child->getSpotOfPoint (px - kx, py - ky);
	  break;
	case DIR_UP:
	  if (y <= py) return child->getSpotOfPoint (px - kx, py - ky);
	  break;
	case DIR_RIGHT:
	  if (x > px) return child->getSpotOfPoint (px - kx, py - ky);
	  break;
	case DIR_DOWN:
	  if (y > py) return child->getSpotOfPoint (px - kx, py - ky);
	  break;
	}
    }

  /* click must be after last child. */
  return child->getSpotOfPoint (px - kx, py - ky);
}


/* this finds the extra space in this slab along its major axis
   (extra meaning the space that the children aren't taking up).
   this is used for centering */
int _ContSlab::find_free_space ()
{
  int cSMaj = 0;
  int i;

  for (i=0; (guint)i<g_list_length (children); i++)
    {
      Slab c = (Slab) getChildAt (i);
      cSMaj += c->getMMajorHigh (style->getDirection ()) +
	c->getMMajorLow (style->getDirection ());
    }

  return
    getMajorLow (style->getDirection ()) +
    getMajorHigh (style->getDirection ()) - cSMaj;
}


/* this changes *x and *y based on the alignment of this's style.
   (low, center, high) */
int _ContSlab::alignmentAdd (int *x, int *y)
{
  /* deal with alignment (chaotic neutral) */
  int add = 0;
  char alignment = style->getAlignment ();
  WPSizePolicy xsp, ysp;
  xsp = style->getXSizePolicy ();
  ysp = style->getYSizePolicy ();

  switch (alignment)
    {
    case ALIGN_LOW: add = 0; break;
    case ALIGN_CENTER: add = find_free_space ()/2; break;
    case ALIGN_HIGH: add = find_free_space (); break;
    }

  switch (style->getDirection ())
    {
    case DIR_LEFT:
      if (x)
	(*x) -= add;
      break;
    case DIR_UP:
      if (y)
	(*y) -= add;
      break;
    case DIR_RIGHT:
      if (x)
	(*x) += add;
      break;
    case DIR_DOWN:
      if (y)
	(*y) += add;
      break;
    }

  /* try to keep children visible */
  if (xsp == SIZE_POLICY_FIXED)
    {
      if (style->getDirection () == DIR_LEFT && left == 0)
	(*x) += right;
      if (style->getDirection () == DIR_RIGHT && right == 0)
	(*x) -= left;
    }
  if (ysp == SIZE_POLICY_FIXED)
    {
      if (style->getDirection () == DIR_UP && up == 0)
	(*x) += down;
      if (style->getDirection () == DIR_DOWN && down == 0)
	(*x) -= up;
    }


  return add;
}


void _ContSlab::draw_margin (int x, int y, GList *views, int /*debug*/)
{
  WPMarginRenderingPolicy mrp = style->getMarginRendering ();

  if (mrp != MARGIN_RENDERING_POLICY_BLANK)
    {
      for (GList *v = views; v != NULL; v = v->next)
	{
	  WPView *g = (WPView *) v->data;
	  GdkColor *c = style->getForeground ();
	  wp_view_set_color (g, c);

	  /***
  
	      0   1               2   3 
	     0+-----------b-----------+
	      |                       |
	      1   +-------f-------+   |
	      |   |               |   |
	      |   |               |   |
	      a   e      x,y      g   c
	      |   |               |   |
	      |   |               |   |
	      |   |               |   |
	      2   +-------h-------+   |
	      |                       |
	     3+-----------d-----------+
  
	  ***/

	  int x0 = x - getMLeft ();
	  int x1 = x - getLeft () - 1;
	  int y0 = y - getMUp ();
	  int y1 = y - getUp () - 1;
	  int x2 = x + getRight () + 1;
	  int x3 = x + getMRight () /*- 1*/;
	  int y2 = y + getDown () + 1;
	  int y3 = y + getMDown (); /*- 1*/;

	  if (mrp == MARGIN_RENDERING_POLICY_OUTLINE ||
	      mrp == MARGIN_RENDERING_POLICY_FILL)
	    {
	      GdkColor *mc = preferences->margin_color;
	      wp_view_set_color (g, mc);

	      if (style->getMarginLeft ())
		/* e */ wp_view_draw_line (g, x1, y1, x1, y2);
	      if (style->getMarginUp ())
		/* f */ wp_view_draw_line (g, x1, y1, x2, y1);
	      if (style->getMarginRight ())
		/* g */ wp_view_draw_line (g, x2, y1, x2, y2);
	      if (style->getMarginDown ())
		/* h */ wp_view_draw_line (g, x1, y2, x2, y2);
	    }

	  if (mrp == MARGIN_RENDERING_POLICY_FILL)
	    {
	      wp_view_set_color (g, c);

	      /*
	      wp_view_fill_rect (g, x0+1, y0+1, (x1-x0)-1, (y3-y0)-2);
	      wp_view_fill_rect (g, x1, y0+1, (x2-x1), (y1-y0)-1);
	      wp_view_fill_rect (g, x2, y0+1, (x3-x2)-1, (y3-y0)-2);
	      wp_view_fill_rect (g, x1, y2, (x2-x1), (y3-y2)-1);
	      */
	      int w = 10;
	      /* XXX check margins... if they are too thin, make w smaller */

	      wp_view_fill_rect (g, x0+1, y0+1, w, (y3-y0)-2);
	      wp_view_fill_rect (g, x0+1, y0+1, (x3-x0)-2, w);
	      wp_view_fill_rect (g, x3-w-1, y0+1, w, (y3-y0)-2);
	      wp_view_fill_rect (g, x0+1, y3-w-1, (x3-x0)-2, w);
	    }


#         if 0
	  /* a */ wp_view_draw_line (g, x0, y0, x0, y3);
	  /* b */ wp_view_draw_line (g, x0, y0, x3, y0);
	  /* c */ wp_view_draw_line (g, x3, y0, x3, y3);
	  /* d */ wp_view_draw_line (g, x0, y3, x3, y3);
	  /* e */ wp_view_draw_line (g, x1, y1, x1, y2);
	  /* f */ wp_view_draw_line (g, x1, y1, x2, y1);
	  /* g */ wp_view_draw_line (g, x2, y1, x2, y2);
	  /* h */ wp_view_draw_line (g, x1, y2, x2, y2);
	  wp_view_draw_line (g, x0, y0, x1, y1);
	  wp_view_draw_line (g, x3, y0, x2, y1);
	  wp_view_draw_line (g, x0, y3, x1, y2);
	  wp_view_draw_line (g, x2, y2, x3, y3);
#         endif 0
	}
    }  
}


/* this recurses to each of the children contained in this
   and draws them */
void _ContSlab::draw (int x, int y, GList *views, int debug)
{
  int num = g_list_length (children);
  Slab child;
  int sx, sy;

  if (debug >= 0)
    {
      printf ("%sdraw this=", tabber (debug));
      dd ();
      printf (" xy=%d,%d\n", x, y);
    }

  /* if this slab isn't visable from any view, bail out early */
  {
    GList *v;
    for (v = views; v != NULL; v = v->next)
      {
	WPView *g = (WPView *) v->data;
	if (slab_overlaps_view (x, y, g))
	  break;
      }
    if (v == NULL) return;
  }

  sx=x;
  sy=y;

  alignmentAdd (&x, &y);

  for (int i=0; i<num; i++)
    {
      child = (Slab) g_list_nth (children, i)->data;

      switch (style->getDirection ())
	{
	case DIR_LEFT:
	  x -= child->getMRight ();
	  break;
	case DIR_UP:
	  y -= child->getMDown ();
	  break;
	case DIR_RIGHT:
	  x += child->getMLeft ();
	  break;
	case DIR_DOWN:
	  y += child->getMUp ();
	  break;
	}

      child->draw (x, y, views, (debug >= 0)?debug+1:-1);

      switch (style->getDirection ())
	{
	case DIR_LEFT: x -= child->getMLeft (); break;
	case DIR_UP: y -= child->getMUp (); break;
	case DIR_RIGHT: x += child->getMRight (); break;
	case DIR_DOWN: y += child->getMDown (); break;
	}
    }

  draw_margin (sx, sy, views, debug);

  for (GList *v = views; v != NULL; v = v->next)
    {
      WPView *g = (WPView *) v->data;

      if (! slab_overlaps_view (x, y, g))
	continue;

      if (xdrawfunc)
	(*xdrawfunc) (g, this, x, y);
    }
}


/* erase this */
void _ContSlab::erase (int x, int y, GList *views)
{
  alignmentAdd (&x, &y);

  for (GList *v = views; v != NULL; v = v->next)
    {
      WPView *g = (WPView *) v->data;
      if (! slab_overlaps_view (x, y, g)) continue;
      wp_view_set_color (g, style->getBackground ());
      wp_view_fill_rect (g, x-left, y-up, left+right, up+down);
    }
}


/**
 * _ContSlab::drawASCII:
 * @doc: document containing this slab
 * @x: the slab's X position
 * @y: the slab's Y position
 * 
 * free what this returns
 * 
 * Return value: a malloced char buffer with ascii text of this slab or NULL
 **/
char *_ContSlab::drawASCII (WPDocument *doc, int x, int y)
{
  /* XXX draw children */

  if (asciidrawfunc)
    return (*asciidrawfunc) (doc, this, x, y);
  return NULL;
}


void _ContSlab::drawPS (WPDocument *doc, GnomePrintContext *pc, int x, int y)
{
  /* XXX draw children */

  if (psdrawfunc)
    (*psdrawfunc) (doc, pc, this, x, y);
}


/* when something changes in a style, every slab that
   uses that style might change size/color/etc.  This
   does a depth first search of the slab tree looking
   for slabs that use the style "s".  Whenever it finds
   one, it calls the overloaded function in wp_slab.C
   which "redoes" the slab.  It recalculates the size,
   erases and redraws this slab. */
int _ContSlab::styleChangeNotify (WPStyle *s, GList *spots,
				  GList *views, int debug)
{
  int ret;

  ret = _Slab::styleChangeNotify (s, spots, views, debug);

  /* g_list_length () might change while this is going on */
  for (unsigned int i=0; i<g_list_length (children); i++)
    {
      int cret;

      Slab child = (Slab) g_list_nth (children, i)->data;
      cret = child->styleChangeNotify (s, spots, views, debug);
      if (cret)
	ret = 1;
    }

  return ret;
}
