
/* for stat */
#include <sys/types.h>
#include <sys/stat.h>
/* for scandir */
#include <dirent.h>


#include <glib.h>
#include <gmodule.h>
#include "wp_plugin.h"
#include "wp_plugin_manager.h"

#ifndef PLUGIN_SPEW
#define PLUGIN_SPEW 0
#endif


GList *WPPluginManager::plugins = NULL;
GHashTable *WPPluginManager::plugins_hash = NULL;


WPPluginManager::~WPPluginManager ()
{
}


WPPluginManager::WPPluginManager ()
{
}


static WPPluginManager* mgr = NULL;

WPPluginManager *
WPPluginManager::thePluginManager ()
{
  if (!mgr)
    mgr = new WPPluginManager ();
  return mgr;
}


void WPPluginManager::scan_for_plugins ()
{
  char buf[1000];
  char *HOME = getenv("HOME");
  char *gpp = getenv("GWP_PLUGIN_PATH");

  /* we look in three places, in this order.
     1) $GWP_PLUGIN_PATH
     2) ~/.gwp/plugins/
     3) <exec_prefix>/lib/gwp/plugins/
     */

  /* I'm adding this so gwp will just *work* if someone builds it
     and runs it in place */
  if (gpp == NULL)
    {
      gpp = strdup ("./plugins/doc/.libs:./plugins/io/.libs");
#     if PLUGIN_SPEW
      fprintf (stderr,
	       "GWP_PLUGIN_PATH not set, using '%s'", gpp);
#     endif
    }

  char *dir = gpp;

  while (dir)
    {
      char *colon = strchr(dir, ':');

      if (colon) *colon = '\0';

      if (strlen(dir) > 0)
	scan_for_plugins_in_directory (dir);

      if (colon)
	{
	  *colon = ':';
	  dir = colon + 1;
	}
      else
	{
	  dir = NULL;
	}
    }

  if (HOME)
    {
      sprintf (buf, "%s%s%s",
	       HOME,
	       HOME[strlen(HOME) - 1] == '/' ? "" : "/",
	       ".gwp/plugins");

      scan_for_plugins_in_directory (buf);
    }
  else
    {
      gnome_error_dialog(_("getenv(\"HOME\") failed."));
    }

  /* XXX - should be exec_prefix somehow... */
  scan_for_plugins_in_directory (DEFAULT_PLUGIN_PATH);
}


static int scandir_dso_select (struct dirent *d)
{
  /*return gnome_dl_is_library_filename (d->d_name);*/
  int len = strlen (d->d_name);

  if (len < 4)
    return 0;

  if (d->d_name[ len-1 ] == 'o' &&
      d->d_name[ len-2 ] == 's' &&
      d->d_name[ len-3 ] == '.')
    return 1;

  return 0;
}


void WPPluginManager::scan_for_plugins_in_directory (char *dir)
{
  struct stat stat_buf;
  struct dirent **namelist;
  int num_names, i;
  
  if (stat(dir, &stat_buf) == -1
      || !(stat_buf.st_mode & S_IFDIR))
    {
#     if PLUGIN_SPEW
      fprintf (stderr,
	       "%s was not a directory, not scanning for plugins there.\n",
	       dir);
#     endif
      return;
    }
  
  if ((num_names = scandir(dir, &namelist, scandir_dso_select, NULL)) == -1)
    {
#     if PLUGIN_SPEW
      fprintf (stderr,
	       "scandir failed on %s, not scanning for plugins there.\n",
	       dir);
#     endif
      return;
    }
  
  for (i = 0; i < num_names; i ++)
    {
      char path[ 2048 ];
      GModule *module;

      /* construct the full path name */
      if (strlen (dir) + strlen (namelist[i]->d_name) + 5 > sizeof (path))
	continue;
      sprintf (path, "%s/%s", dir, namelist[i]->d_name);

#     if PLUGIN_SPEW
      printf ("checking for plugin in '%s'\n", path);
#     endif

      /* attempt to load the module */
      module = g_module_open (path, G_MODULE_BIND_LAZY);
      /*
	G_MODULE_BIND_LAZY	= 1 << 0,
	G_MODULE_BIND_MASK	= 0x01
      */


      /* check for error */
      if (!module)
	{
#         if PLUGIN_SPEW
	  fprintf (stderr, "module '%s' -- %s\n", path, g_module_error ());
#         endif
	  continue;
	}

      /* see if it is intended for use by gwp */
      PluginFillVTable plugin_fill_vtable;
      if (! g_module_symbol (module,
			     "wp_plugin_fill_vtable",
			     (gpointer *) &plugin_fill_vtable))
	{
#         if PLUGIN_SPEW
	  fprintf (stderr,
		   "module '%s' -- can't find wp_plugin_fill_vtable\n", path);
#         endif
	  continue;
	}

      /* get the plugins vtable entries */
      PluginVTable *pvt = (PluginVTable *) calloc (1, sizeof (PluginVTable));

      if (plugin_fill_vtable (pvt))
	{
#         if PLUGIN_SPEW
	  fprintf (stderr,
		   "module '%s' -- plugin_fill_vtable failed\n", path);
#         endif
	  continue;
	}

      /* create a new C++ object for the plugin */
      WPPlugin *plugin = new WPPlugin (pvt);

      /* initialize this plugin */

      /* add this plugin to the name hash */
      if (!plugins_hash)
	plugins_hash = g_hash_table_new (g_str_hash, g_str_equal);

      char *hash_entry = plugin->getUIName ();
      int found = g_hash_table_lookup (plugins_hash, hash_entry) != NULL;
      if (!found && hash_entry)
	{
#         if PLUGIN_SPEW
	  printf ("+ adding plugin\n");
#         endif
	  g_hash_table_insert(plugins_hash, hash_entry, plugin);
	  plugins = g_list_append (plugins, plugin);
	}
      else
	{
#         if PLUGIN_SPEW
	  printf ("+ plugin for this type already exists.  "
		  "deleting prospective plugin.\n");
#         endif
	  delete plugin;
	}
    }

  /* Free up the memory returned by `scandir'. */
  while (num_names > 0)
    free (namelist[--num_names]);
  free (namelist);
}


WPPlugin *WPPluginManager::get_plugin_by_name (char *name)
{
  g_return_val_if_fail (name, NULL);

  for (GList *pp = plugins; pp; pp=pp->next)
    {
      WPPlugin *plugin = (WPPlugin *) pp->data;
      if (strcmp (name, plugin->getUIName ()) == 0)
	return plugin;
    }

  return NULL;
}
